@@ -1,4 +1,5 @@
-/*Package seq defines a *Seq* type, and provides some basic operations of sequence,
+/*
+Package seq defines a *Seq* type, and provides some basic operations of sequence,
 like validation of DNA/RNA/Protein sequence and getting reverse complement sequence.
 
 This package was inspired by
@@ -8,7 +9,6 @@ IUPAC nucleotide code: ACGTURYSWKMBDHVN
 
 http://droog.gs.washington.edu/parc/images/iupac.html
 
-
 	code	base	Complement
 	A	A	T
 	C	C	G
@@ -65,13 +65,12 @@ IUPAC amino acid code
 
 Reference:
 
-	1. http://www.bioinformatics.org/sms/iupac.html
-	2. http://www.dnabaser.com/articles/IUPAC%20ambiguity%20codes.html
-	3. http://www.bioinformatics.org/sms2/iupac.html
-	4. http://www.matrixscience.com/blog/non-standard-amino-acid-residues.html
-	5. http://www.sbcs.qmul.ac.uk/iupac/AminoAcid/A2021.html#AA21
-	6. https://en.wikipedia.org/wiki/Amino_acid
-
+ 1. http://www.bioinformatics.org/sms/iupac.html
+ 2. http://www.dnabaser.com/articles/IUPAC%20ambiguity%20codes.html
+ 3. http://www.bioinformatics.org/sms2/iupac.html
+ 4. http://www.matrixscience.com/blog/non-standard-amino-acid-residues.html
+ 5. http://www.sbcs.qmul.ac.uk/iupac/AminoAcid/A2021.html#AA21
+ 6. https://en.wikipedia.org/wiki/Amino_acid
 */
 package seq
 
@@ -81,10 +80,12 @@ import (
 	"runtime"
 	"sync"
 
+	"github.com/elliotwutingfeng/asciiset"
 	"github.com/shenwei356/util/byteutil"
 )
 
-/*Alphabet could be defined. Attention that,
+/*
+Alphabet could be defined. Attention that,
 **the letters are case sensitive**.
 
 For example, DNA:
@@ -95,7 +96,6 @@ For example, DNA:
 		[]byte("tgcaTGCA"),
 		[]byte(" -"),
 		[]byte("nN"))
-
 */
 type Alphabet struct {
 	t         string
@@ -223,9 +223,6 @@ func (a *Alphabet) IsValidLetter(b byte) bool {
 // needed to  parallelly checking sequence
 var ValidSeqLengthThreshold = 10000
 
-// ValidateWholeSeq is used to determin whether validate all bases of a seq
-var ValidateWholeSeq = true
-
 // ValidSeqThreads is the threads number of parallelly checking sequence
 var ValidSeqThreads = runtime.NumCPU()
 
@@ -254,15 +251,13 @@ func (a *Alphabet) IsValid(s []byte) error {
 		return nil
 	}
 
-	if ValidateWholeSeq || ValidSeqThreads == 0 {
-		ValidSeqThreads = len(s)
-	}
-	chunkSize, start, end := int(l/ValidSeqThreads), 0, 0
+	chunkSize, start, end := int(l/ValidSeqThreads)+1, 0, 0
 
 	var wg sync.WaitGroup
 	tokens := make(chan int, ValidSeqThreads)
 	ch := make(chan seqCheckStatus, ValidSeqThreads)
 	done := make(chan struct{})
+	var once sync.Once
 	finished := false
 	for i := 0; i < ValidSeqThreads; i++ {
 		start = i * chunkSize
@@ -282,7 +277,7 @@ func (a *Alphabet) IsValid(s []byte) error {
 			case <-done:
 				if !finished {
 					finished = true
-					close(ch)
+					// close(ch)
 					return
 				}
 			default:
@@ -294,7 +289,9 @@ func (a *Alphabet) IsValid(s []byte) error {
 				j = int(s[i])
 				if j >= len(a.pairLetters) || a.pairLetters[j] == 0 {
 					ch <- seqCheckStatus{fmt.Errorf("seq: invalid %s lebtter: %s at %d", a, []byte{s[i]}, i)}
-					close(done)
+					once.Do(func() {
+						close(done)
+					})
 					return
 				}
 			}
@@ -327,15 +324,15 @@ func (a *Alphabet) PairLetter(b byte) (byte, error) {
 	return p, nil
 }
 
-/*Four types of alphabets are pre-defined:
-
-  DNA           Deoxyribonucleotide code
-  DNAredundant  DNA + Ambiguity Codes
-  RNA           Oxyribonucleotide code
-  RNAredundant  RNA + Ambiguity Codes
-  Protein       Amino Acide single-letter Code
-  Unlimit       Self-defined, including all 26 English letters
+/*
+Four types of alphabets are pre-defined:
 
+	DNA           Deoxyribonucleotide code
+	DNAredundant  DNA + Ambiguity Codes
+	RNA           Oxyribonucleotide code
+	RNAredundant  RNA + Ambiguity Codes
+	Protein       Amino Acide single-letter Code
+	Unlimit       Self-defined, including all 26 English letters
 */
 var (
 	DNA          *Alphabet
@@ -345,11 +342,11 @@ var (
 	Protein      *Alphabet
 	Unlimit      *Alphabet
 
-	abProtein      map[byte]bool
-	abDNAredundant map[byte]bool
-	abDNA          map[byte]bool
-	abRNAredundant map[byte]bool
-	abRNA          map[byte]bool
+	abProtein      asciiset.ASCIISet
+	abDNAredundant asciiset.ASCIISet
+	abDNA          asciiset.ASCIISet
+	abRNAredundant asciiset.ASCIISet
+	abRNA          asciiset.ASCIISet
 )
 
 func init() {
@@ -417,7 +414,7 @@ func GuessAlphabet(seqs []byte) *Alphabet {
 	if len(seqs) == 0 {
 		return Unlimit
 	}
-	var alphabetMap map[byte]bool
+	var alphabetMap asciiset.ASCIISet
 	if AlphabetGuessSeqLengthThreshold == 0 || len(seqs) <= AlphabetGuessSeqLengthThreshold {
 		alphabetMap = slice2map(byteutil.Alphabet(seqs))
 	} else { // reduce guessing time
@@ -454,19 +451,14 @@ func GuessAlphabetLessConservatively(seqs []byte) *Alphabet {
 	return ab
 }
 
-func isSubset(query, subject map[byte]bool) bool {
-	for b := range query {
-		if _, ok := subject[b]; !ok {
-			return false
-		}
-	}
-	return true
+// isSubset returns true if query is a subset of subject
+func isSubset(query, subject asciiset.ASCIISet) bool {
+	// A ⊆ B iff (A ∪ B) = B
+	union := query.Union(subject)
+	return union.Equals(subject)
 }
 
-func slice2map(s []byte) map[byte]bool {
-	m := make(map[byte]bool)
-	for _, b := range s {
-		m[b] = true
-	}
+func slice2map(s []byte) asciiset.ASCIISet {
+	m, _ := asciiset.MakeASCIISet(string(s))
 	return m
 }
